#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <sys/errno.h>
#include "netport.h"

/* client port access stubs */

static int net_fd = -1;

/* -1 on failure, file descriptor of socket to server on success */
static int connect_to_server(char * server_name, int port) {
    struct hostent *server_host;    /* info about the server when we find it */
    struct sockaddr_in my_addr;
    struct sockaddr_in serv_addr;
    int fd;

    /* create a TCP/IP socket */
    if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        perror("netport: failed to create socket");
        return -1;
    }

#ifdef DEBUG
    fprintf(stderr,"netport: socket created\n");
#endif

    /* fill in our local information */
    memset((char *)&my_addr, 0, sizeof(my_addr));
    my_addr.sin_family = AF_INET;
    my_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    my_addr.sin_port = htons(0); /* arbitrary port */

    /* bind! */
    if (bind(fd, (struct sockaddr *)&my_addr, sizeof(my_addr)) < 0) {
        perror("netport: bind failed");
        close(fd);
        return -1;
    }

#ifdef DEBUG
    fprintf(stderr,"netport: bind succeeded\n");
#endif

    /* fill in the server's information */
    memset((char *)&serv_addr, 0, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_port = htons(port);

    /* look up the address of the server */
    server_host = gethostbyname(server_name);
    if (!server_host) {
        fprintf(stderr,"netport: failed to look up server address\n");
        close(fd);
        return -1;
    }

#ifdef DEBUG
    fprintf(stderr,"netport: gethostbyname succeeded\n");
#endif

    /* copy the server's address */
    memcpy((caddr_t)&serv_addr.sin_addr, server_host->h_addr_list[0], server_host->h_length);

    /* connect! */
    if (connect(fd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) {
        perror("netport: connect failed");
        close(fd);
        return -1;
    }

#ifdef DEBUG
    fprintf(stderr,"netport: connect succeeded\n");
#endif

    /* send signature */
    if (write_or_fail(fd, (unsigned char*)NETPORT_HANDSHAKE, strlen(NETPORT_HANDSHAKE))) {
        perror("netport: failed sending handshake");
        close(fd);
        return -1;
    }

#ifdef DEBUG
    fprintf(stderr,"netport: handshake sent\n");
#endif

    /* all seems well! */
    return fd;
}

static int get_net_fd(void)
{
    if (net_fd == -1) {
        char * server;
        server = getenv("NETPORT_SERVER");
        if (server == NULL) {
            fprintf(stderr,"netport: NETPORT_SERVER environment variable not set!\n");
            return -1;
        }
#if DEBUG
        fprintf(stderr,"netport: attempting to connect to %s\n",server);
#endif

        if ((net_fd = connect_to_server(server,NETPORT_PORT)) == -1) {
            fprintf(stderr,"netport: failed to connect to %s\n",server);
            return -1;
        }
    }

    return net_fd;
}

void netport_end(void) {
    const unsigned char command = cmd_end;
    if (net_fd == -1) return;
    
    if (write_or_fail(net_fd,&command,1)) {
        perror("netport: failed writing cmd_end");
    }
    close(net_fd);
    net_fd = -1;
}

void netport_nop(void) {
    const unsigned char command = cmd_nop;
    int fd;
    unsigned char response;
    
    if ((fd=get_net_fd()) < 0) return;
    if (write_or_fail(fd,&command,1)) {
        perror("netport: failed writing cmd_nop");
        return;
    }
    if (read_or_fail(fd,&response,1)) {
        perror("netport: failed reading cmd_nop response");
    }

    if (response != 0) {
        fprintf(stderr,"netport: received invalid response to cmd_nop");
    }
}


void outw (unsigned short int value, unsigned short int port) {
    const unsigned char command = cmd_outw;
    int fd;
#ifdef DEBUG
    fprintf(stderr,"netport: %4hx->[%4hx]\n",value,port);
#endif

    if ((fd=get_net_fd()) < 0) return;
    value = htons(value);
    port = htons(port);
    if (write_or_fail(fd,&command,1) ||
            write_or_fail(fd,(unsigned char *)&port,2) ||
            write_or_fail(fd,(unsigned char *)&value,2)) {
        perror("netport: failed writing cmd_outw");
        return;
    }
}

unsigned short int inw (unsigned short int port)
{
    const unsigned char command = cmd_inw;
    int fd;
    unsigned short value;
#ifdef DEBUG
    fprintf(stderr,"netport: ????" "<-[%4hx]\n",port);
#endif
    
    if ((fd=get_net_fd()) < 0) return 0;
    port = htons(port);
    if (write_or_fail(fd,&command,1) ||
            write_or_fail(fd,(unsigned char *)&port,2)) {
        perror("netport: failed writing cmd_inw");
        return 0;
    }

    if (read_or_fail(fd,(unsigned char *)&value,2)) {
        perror("netport: failed reading cmd_inw response");
        return 0;
    }

    value = ntohs(value);

    return value;
}

void outb (unsigned char value, unsigned short int port)
{
    const unsigned char command = cmd_outb;
    int fd;
#ifdef DEBUG
    fprintf(stderr,"netport: %2hhx  ->[%4hx]\n",value,port);
#endif
    

    if ((fd=get_net_fd()) < 0) return;
    port = htons(port);
    if (write_or_fail(fd,&command,1) ||
            write_or_fail(fd,(unsigned char *)&port,2) ||
            write_or_fail(fd,&value,1)) {
        perror("netport: failed writing cmd_outb");
        return;
    }
}

unsigned char inb (unsigned short int port)
{
    const unsigned char command = cmd_inb;
    int fd;
    unsigned char value;
#ifdef DEBUG
    fprintf(stderr,"netport: ??  <-[%4hx]\n",port);
#endif

    if ((fd=get_net_fd()) < 0) return 0;
    port = htons(port);
    if (write_or_fail(fd,&command,1) ||
            write_or_fail(fd,(unsigned char *)&port,2)) {
        perror("netport: failed writing cmd_inb");
        return 0;
    }

    if (read_or_fail(fd,&value,1)) {
        perror("netport: failed reading cmd_inb response");
        return 0;
    }

    return value;
}

int ioperm(unsigned long from, unsigned long num, int turn_on)
{
    const unsigned char command = cmd_ioperm;
    int fd;
    unsigned char turn_on_char;
    unsigned char response;
    unsigned long error;
#ifdef DEBUG
    fprintf(stderr,"netport: ioperm(%lx,%lx,%s)\n",from,num,(turn_on?"on":"off"));
#endif

    if ((fd = get_net_fd()) < 0) return -1;
    from = htonl(from);
    num = htonl(num);
    turn_on_char = (turn_on != 0);

    if (write_or_fail(fd,&command,1) ||
            write_or_fail(fd,(unsigned char *)&from,4) ||
            write_or_fail(fd,(unsigned char *)&num,4) ||
            write_or_fail(fd,&turn_on_char,1)) {
        perror("netport: failed writing cmd_ioperm");
        return -1;
    }

    if (read_or_fail(fd,&response,1) ||
            read_or_fail(fd,(unsigned char *)&error,4)) {
        perror("netport: failed reading cmd_ioperm response");
        return 0;
    }

    error = ntohl(error);

    if (response != 0)
        errno = error;
    return (response==0)?0:-1;
}
