#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/errno.h>
#include <unistd.h>
#include <sys/io.h>
#include "netport.h"

/* server process (with actual access to the ports) */

void service_request(int client_fd)
{
    unsigned char buf[10];
    int connection_alive;

    if (read_or_fail(client_fd,buf,strlen(NETPORT_HANDSHAKE))) {
        perror("error reading handshake");
        return;
    }

    if (memcmp(buf,NETPORT_HANDSHAKE,strlen(NETPORT_HANDSHAKE))) {
        fprintf(stderr,"invalid handshake\n");
        return;
    }

#ifdef DEBUG
    fprintf(stderr,"handshake ok\n");
#endif

    connection_alive = 1;
    while (connection_alive) {
        unsigned char command;
        if (read_or_fail(client_fd,&command,1)) {
            perror("error reading command number");
            return;
        }

        switch (command) {
            case cmd_nop:
#ifdef DEBUG
                fprintf(stderr,"cmd_nop\n");
#endif
                buf[0] = 0;
                if (write_or_fail(client_fd,buf,1)) {
                    perror("error writing cmd_nop response");
                    return;
                }
                break;

            case cmd_end:
#ifdef DEBUG
                fprintf(stderr,"cmd_end\n");
#endif
                connection_alive = 0;
                break;

            case cmd_outw:
#ifdef DEBUG
                fprintf(stderr,"cmd_outw\n");
#endif
                {
                    unsigned short port, value;
                    if (read_or_fail(client_fd,(unsigned char *)&port,2) ||
                            read_or_fail(client_fd,(unsigned char *)&value,2)) {
                        perror("error reading cmd_outw body");
                        return;
                    }
                    port = ntohs(port);
                    value = ntohs(value);

                    outw(value,port);
#ifdef DEBUG
                    fprintf(stderr,"%4hx->[%4hx]\n",value,port);
#endif
                }
                break;

            case cmd_inw:
#ifdef DEBUG
                fprintf(stderr,"cmd_inw\n");
#endif
                {
                    unsigned short port, value;
                    if (read_or_fail(client_fd,(unsigned char *)&port,2)) {
                        perror("error reading cmd_inw body");
                        return;
                    }
                    port = ntohs(port);

                    value = htons(inw(port));
#ifdef DEBUG
                    fprintf(stderr,"%4hx<-[%4hx]\n",value,port);
#endif
                    if (write_or_fail(client_fd,(unsigned char *)&value,2)) {
                        perror("error sending cmd_inw response");
                        return;
                    }
                }
                break;

            case cmd_outb:
#ifdef DEBUG
                fprintf(stderr,"cmd_outb\n");
#endif
                {
                    unsigned short port;
                    unsigned char value;
                    if (read_or_fail(client_fd,(unsigned char *)&port,2) ||
                            read_or_fail(client_fd,&value,1)) {
                        perror("error reading cmd_outb body");
                        return;
                    }
                    port = ntohs(port);

                    outb(value,port);
#ifdef DEBUG
                    fprintf(stderr,"%2hhx  ->[%4hx]\n",value,port);
#endif
                }
                break;

            case cmd_inb:
#ifdef DEBUG
                fprintf(stderr,"cmd_inb\n");
#endif
                {
                    unsigned short port;
                    unsigned char value;
                    if (read_or_fail(client_fd,(unsigned char *)&port,2)) {
                        perror("error reading cmd_inb body");
                        return;
                    }
                    port = ntohs(port);

                    value = inb(port);
#ifdef DEBUG
                    fprintf(stderr,"%2hhx  <-[%4hx]\n",value,port);
#endif

                    if (write_or_fail(client_fd,&value,1)) {
                        perror("error writing cmd_inb response");
                        return;
                    }
                }
                break;

            case cmd_ioperm:
#ifdef DEBUG
                fprintf(stderr,"cmd_ioperm\n");
#endif
                {
                    unsigned long startport;
                    unsigned long numaddr;
                    unsigned char enable;
                    unsigned char result;
                    unsigned long error;

                    if (read_or_fail(client_fd,(unsigned char *)&startport,4) ||
                            read_or_fail(client_fd,(unsigned char *)&numaddr,4) ||
                            read_or_fail(client_fd,&enable,1)) {
                        perror("error reading cmd_ioperm body");
                        return;
                    }
                    startport = ntohl(startport);
                    numaddr = ntohl(numaddr);

                    error = 0;
		    result = ioperm(startport,numaddr,enable)==-1?1:0;
		    if (!result) error=errno;
#ifdef DEBUG
                    fprintf(stderr,"ioperm(%lx,%lx,%s)\n",startport,numaddr,(enable?"on":"off"));
#endif

                    error = htonl(error);

                    if (write_or_fail(client_fd,&result,1) ||
                            write_or_fail(client_fd,(unsigned char *)&error,4)) {
                        perror("error sending cmd_ioperm response");
                        return;
                    }
                }
                break;

            default:
                fprintf(stderr,"invalid command number %d\n",command);
                return;
        }
    }
}

void serve(int port)
{
    int listening_socket;
    struct sockaddr_in my_addr;
    struct sockaddr_in client_addr;
    int sockoptval = 1;

    /* get a TCP socket */
    if ((listening_socket = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        perror("creating listening socket failed");
        exit(1);
    }

    /* set SO_REUSEADDR so we can immediately reuse the port on exit */
    setsockopt(listening_socket, SOL_SOCKET, SO_REUSEADDR, &sockoptval, sizeof(int));

    /* prepare for bind: fill in my end */
    memset(&my_addr, 0, sizeof(my_addr));
    my_addr.sin_family = AF_INET;
    my_addr.sin_port = htons(port);
    my_addr.sin_addr.s_addr = htonl(INADDR_ANY);

    /* bind! */
    if (bind(listening_socket, (struct sockaddr *) &my_addr, sizeof(my_addr)) < 0) {
        perror("bind failed");
        exit(1);
    }

    /* now that we're bound start listening */
    /* a queue of 1 is fine as we only intend this for personal use */
    if (listen(listening_socket, 1) < 0) {
        perror("listen failed");
        exit(1);
    }

    for (;;) {
        int request_socket;
        unsigned int client_addr_len = sizeof(client_addr);
#ifdef DEBUG
        fprintf(stderr,"listening...\n");
#endif
        while ((request_socket = accept(listening_socket,
                        (struct sockaddr *) &client_addr, &client_addr_len)) < 0) {
            if (errno != EINTR) {
                perror("accept failed");
            }
        }

#ifdef DEBUG
        fprintf(stderr,"servicing request\n");
#endif
        service_request(request_socket);
#ifdef DEBUG
        fprintf(stderr,"request finished (one way or another)\n");
#endif
        close(request_socket);
    }
}

int main(void)
{

    serve(NETPORT_PORT);

    return 0;
}
